/* New simplified depmod without backwards compat stuff and not
   requiring ksyms.

   (C) 2002 Rusty Russell IBM Corporation
 */
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <elf.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <dirent.h>
#include <sys/utsname.h>
#include <sys/mman.h>

#include "depmod.h"
#include "moduleops.h"
#include "tables.h"

#include "testing.h"

/* I hate strcmp. */
#define streq(a,b) (strcmp((a),(b)) == 0)

void fatal(const char *fmt, ...)
{
	va_list arglist;

	fprintf(stderr, "FATAL: ");

	va_start(arglist, fmt);
	vfprintf(stderr, fmt, arglist);
	va_end(arglist);

	exit(1);
}

void warn(const char *fmt, ...)
{
	va_list arglist;

	fprintf(stderr, "WARNING: ");

	va_start(arglist, fmt);
	vfprintf(stderr, fmt, arglist);
	va_end(arglist);
}

void *do_nofail(void *ptr, const char *file, int line, const char *expr)
{
	if (!ptr) {
		fatal("Memory allocation failure %s line %d: %s.\n",
		      file, line, expr);
	}
	return ptr;
}

#define SYMBOL_HASH_SIZE 1024
struct symbol
{
	struct symbol *next;
	struct module *owner;
	char name[0];
};

static struct symbol *symbolhash[SYMBOL_HASH_SIZE];

/* This is based on the hash agorithm from gdbm, via tdb */
static inline unsigned int tdb_hash(const char *name)
{
	unsigned value;	/* Used to compute the hash value.  */
	unsigned   i;	/* Used to cycle through random values. */

	/* Set the initial value from the key size. */
	for (value = 0x238F13AF * strlen(name), i=0; name[i]; i++)
		value = (value + (((unsigned char *)name)[i] << (i*5 % 24)));

	return (1103515243 * value + 12345);
}

void add_symbol(const char *name, struct module *owner)
{
	unsigned int hash;
	struct symbol *new = NOFAIL(malloc(sizeof *new + strlen(name) + 1));

	new->owner = owner;
	strcpy(new->name, name);

	hash = tdb_hash(name) % SYMBOL_HASH_SIZE;
	new->next = symbolhash[hash];
	symbolhash[hash] = new;
}

static int print_unknown;

struct module *find_symbol(const char *name, const char *modname)
{
	struct symbol *s;

	for (s = symbolhash[tdb_hash(name) % SYMBOL_HASH_SIZE]; s; s=s->next) {
		if (streq(s->name, name))
			return s->owner;
	}

	/* Ignore __start_* and __stop_* like kernel might. */
	if (print_unknown
	    && (strncmp(name, "__start_", strlen("__start_")) != 0
		|| strncmp(name, "__stop_", strlen("__stop_")) != 0))
		warn("%s needs unknown symbol %s\n", modname, name);

	return NULL;
}

void add_dep(struct module *mod, struct module *depends_on)
{
	unsigned int i;

	for (i = 0; i < mod->num_deps; i++)
		if (mod->deps[i] == depends_on)
			return;

	mod->deps = realloc(mod->deps, sizeof(mod->deps[0])*(mod->num_deps+1));
	mod->deps[mod->num_deps++] = depends_on;
}

static void load_system_map(const char *filename)
{
	FILE *system_map;
	char line[10240];

	system_map = fopen(filename, "r");
	if (!system_map)
		fatal("Could not open '%s': %s\n", filename, strerror(errno));

	/* eg. c0294200 R __ksymtab_devfs_alloc_devnum */
	while (fgets(line, sizeof(line)-1, system_map)) {
		char *ptr;

		/* Snip \n */
		ptr = strchr(line, '\n');
		*ptr = '\0';

		ptr = strchr(line, ' ');
		if (ptr)
			ptr = strchr(ptr + 1, ' ');
		if (ptr
		    && strncmp(ptr+1, "__ksymtab_", strlen("__ksymtab_")) == 0)
			add_symbol(ptr+1+strlen("__ksymtab_"), NULL);
	}

	/* __this_module is magic inserted by kernel loader. */
	add_symbol("__this_module", NULL);
}

static struct option options[] = { { "all", 0, NULL, 'a' },
				   { "quick", 0, NULL, 'A' },
				   { "basedir", 1, NULL, 'b' },
				   { "errsyms", 0, NULL, 'e' },
				   { "filesyms", 0, NULL, 'F' },
				   { "help", 0, NULL, 'h' },
				   { "show", 0, NULL, 'n' },
				   { "quiet", 0, NULL, 'q' },
				   { "root", 0, NULL, 'r' },
				   { "unresolved-error", 0, NULL, 'u' },
				   { "verbose", 0, NULL, 'v' },
				   { "version", 0, NULL, 'V' },
				   { NULL, 0, NULL, 0 } };

/* Version number or module name?  Don't assume extension. */
static int is_version_number(const char *version)
{
	unsigned int dummy;

	return (sscanf(version, "%u.%u.%u", &dummy, &dummy, &dummy) == 3);
}

static int old_module_version(const char *version)
{
	/* Expect three part version. */
	unsigned int major, sub, minor;

	sscanf(version, "%u.%u.%u", &major, &sub, &minor);

	if (major > 2) return 0;
	if (major < 2) return 1;

	/* 2.x */
	if (sub > 5) return 0;
	if (sub < 5) return 1;

	/* 2.5.x */
	if (minor >= 47) return 0;
	return 1;
}

static void exec_old_depmod(char *argv[])
{
	char *sep;
	char pathname[strlen(argv[0])+1];
	char oldname[strlen("depmod") + strlen(argv[0]) + sizeof(".old")];

	memset(pathname, 0, strlen(argv[0])+1);
	sep = strrchr(argv[0], '/');
	if (sep)
		memcpy(pathname, argv[0], sep - argv[0]+1);
	sprintf(oldname, "%s%s.old", pathname, "depmod");

	/* Recursion detection: we need an env var since we can't
	   change argv[0] (as older modutils uses it to determine
	   behavior). */
	if (getenv("MODULE_RECURSE"))
		return;
	setenv("MODULE_RECURSE", "y", 0);

	execvp(oldname, argv);
	fprintf(stderr,
		"Version requires old depmod, but couldn't run %s: %s\n",
		oldname, strerror(errno));
	exit(2);
}

static void print_usage(const char *name)
{
	fprintf(stderr,
	"%s " VERSION " -- part of " PACKAGE "\n"
	"%s -[aA] [-n -e -v -q -V -r -u]\n"
	"      [-b basedirectory] [forced_version]\n"
	"depmod [-n -e -v -q -r -u] [-F kernelsyms] module1.o module2.o ...\n"
	"If no arguments (except options) are given, \"depmod -a\" is assumed\n"
	"\n"
	"depmod will output a dependancy list suitable for the modprobe utility.\n"
	"\n"
	"\n"
	"Options:\n"
	"\t-a, --all            Probe all modules\n"
	"\t-n, --show           Write the dependency file on stdout only\n"
	"\t-V, --version        Print the release version\n"
	"\t-h, --help           Print this usage message\n"
	"\n"
	"The following options are useful for people managing distributions:\n"
	"\t-b basedirectory\n"
	"\t    --basedir basedirectory    Use an image of a module tree.\n"
	"\t-F kernelsyms\n"
	"\t    --filesyms kernelsyms      Use the file instead of the\n"
	"\t                               current kernel symbols.\n",
	"depmod", "depmod");
}

static int ends_in(const char *name, const char *ext)
{
	unsigned int namelen, extlen;

	/* Grab lengths */
	namelen = strlen(name);
	extlen = strlen(ext);

	if (namelen < extlen) return 0;

	if (streq(name + namelen - extlen, ext))
		return 1;
	return 0;
}

static struct module *grab_module(const char *dirname,
				  const char *filename,
				  struct module *next)
{
	struct stat buf;
	struct module *new;
	int fd;

	new = NOFAIL(malloc(sizeof(*new)
			    + strlen(dirname) + 1 + strlen(filename) + 1));
	sprintf(new->pathname, "%s/%s", dirname, filename);
	INIT_LIST_HEAD(&new->dep_list);

	fd = open(new->pathname, O_RDONLY, 0);
	if (fd < 0) {
		warn("Can't read module %s: %s\n",
		     new->pathname, strerror(errno));
		goto fail;
	}
	fstat(fd, &buf);
	new->mmap = mmap(0, buf.st_size, PROT_READ, MAP_PRIVATE, fd, 0);
	if (new->mmap == MAP_FAILED) {
		warn("Can't map module %s: %s\n",
		     new->pathname, strerror(errno));
		goto fail;
	}

	/* "\177ELF" <byte> where byte = 001 for 32-bit, 002 for 64 */
	if (memcmp(new->mmap, ELFMAG, SELFMAG) != 0) {
		warn("Module %s is not an elf object\n", new->pathname);
		goto fail;
	}

	switch (((char *)new->mmap)[EI_CLASS]) {
	case ELFCLASS32:
		new->ops = &mod_ops32;
		break;
	case ELFCLASS64:
		new->ops = &mod_ops64;
		break;
	default:
		warn("Module %s has elf unknown identifier %i\n",
		     new->pathname, ((char *)new->mmap)[EI_CLASS]);
		goto fail;
	}
	new->ops->load_symbols(new);
	new->ops->fetch_tables(new);

	new->next = next;
	close(fd);
	return new;

 fail:
	munmap(new->mmap, buf.st_size);
	close(fd);
	free(new);
	return next;
}

static char *basename(const char *name)
{
	const char *base = strrchr(name, '/');
	if (base) return (char *)base + 1;
	return (char *)name;
}

struct module_traverse
{
	struct module_traverse *prev;
	struct module *mod;
};

/* Assume we are doing all the modules, so only report each loop once. */
static void report_loop(const struct module *start,
			const struct module_traverse *traverse)
{
	const struct module_traverse *i;

	/* Is start in the loop?  If not, don't report now. eg. a
	   depends on b which depends on c which depends on b.  Don't
	   report when generating depends for a. */
	if (traverse->mod != start)
		return;

	/* Check that start is least alphabetically.  eg.  a depends
	   on b depends on a will get reported for a, not b.  */
	for (i = traverse->prev; i->prev; i = i->prev) {
		if (strcmp(start->pathname, i->mod->pathname) > 0)
			return;
	}

	warn("Loop detected: %s ", start->pathname);
	for (i = traverse->prev; i->prev; i = i->prev)
		fprintf(stderr, "needs %s ", basename(i->mod->pathname));
	fprintf(stderr, "which needs %s again!\n", basename(start->pathname));
}

static int in_loop(struct module *mod, const struct module_traverse *traverse)
{
	const struct module_traverse *i;

	for (i = traverse; i; i = i->prev) {
		if (i->mod == mod)
			return 1;
	}
	return 0;
}

/* Creates a dependency list: returns 0 if a loop was found. */
static int create_dep_list(struct module *start,
			   struct module *mod,
			   struct module_traverse *prev)
{
	struct module_traverse traverse = { .prev = prev, .mod = mod };
	unsigned int i;

	if (in_loop(mod, prev)) {
		report_loop(start, &traverse);
		return 0;
	}

	for (i = 0; i < mod->num_deps; i++) {
		/* If it was previously depended on, move it to the
		   tail.  ie. if a needs b and c, and c needs b, we
		   must order b after c. */
		list_del(&mod->deps[i]->dep_list);
		list_add_tail(&mod->deps[i]->dep_list, &start->dep_list);
		if (!create_dep_list(start, mod->deps[i], &traverse))
			return 0;
	}
	return 1;
}

static void output_deps(struct module *modules,
			unsigned int skipchars,
			FILE *out,
			int verbose)
{
	struct module *i;

	for (i = modules; i; i = i->next)
		i->ops->calculate_deps(i, verbose);

	/* Now dump them out. */
	for (i = modules; i; i = i->next) {
		if (create_dep_list(i, i, NULL)) {
			struct list_head *j, *tmp;

			fprintf(out, "%s:", i->pathname + skipchars);
			list_for_each_safe(j, tmp, &i->dep_list) {
				struct module *dep
					= list_entry(j,struct module,dep_list);
				fprintf(out, " %s", dep->pathname + skipchars);
				list_del_init(j);
			}
			fprintf(out, "\n");
		} else {
			warn("Module %s ignored, due to loop\n",
			     i->pathname + skipchars);
		}
	}
}

static int smells_like_module(const char *name)
{
	if (ends_in(name,".o") || ends_in(name,".klm") || ends_in(name,".ko"))
		return 1;
	return 0;
}

static struct module *grab_dir(const char *dirname, struct module *next)
{
	DIR *dir;
	struct dirent *dirent;

	dir = opendir(dirname);
	if (!dir) {
		warn("Couldn't open directory %s: %s\n",
		     dirname, strerror(errno));
		return next;
	}

	while ((dirent = readdir(dir)) != NULL) {
		/* Several competing extensions here.  .klm, .ko, .o */
		if (smells_like_module(dirent->d_name))
			next = grab_module(dirname, dirent->d_name, next);
		else if (!streq(dirent->d_name, ".")
			 && !streq(dirent->d_name, "..")) {
			struct stat st;

			char subdir[strlen(dirname) + 1
				   + strlen(dirent->d_name) + 1];
			sprintf(subdir, "%s/%s", dirname, dirent->d_name);
			if (lstat(subdir, &st) != 0)
				warn("Couldn't stat %s: %s\n", subdir,
				     strerror(errno));
			else if (S_ISDIR(st.st_mode))
				next = grab_dir(subdir, next);
		}
	}
	return next;
}

int main(int argc, char *argv[])
{
	int opt, all = 0, verbose = 0, bad_option = 0;
	unsigned int skipchars = 0;
	FILE *depout = NULL, *pciout, *usbout, *ccwout;
	char *basedir = "", *dirname, *version, *system_map = NULL;
	struct module *list = NULL;

	/* Don't print out any errors just yet, we might want to exec
           backwards compat version. */
	opterr = 0;
	while ((opt = getopt_long(argc, argv, "ab:ArehnqruvVF:", options, NULL))
	       != -1) {
		switch (opt) {
		case 'a':
			all = 1;
			break;
		case 'b':
			basedir = optarg;
			skipchars = strlen(basedir);
			break;
		case 'A':
			all = 1;
			break;
		case 'F':
			system_map = optarg;
			break;
		case 'e':
			print_unknown = 1;
			break;
		case 'v':
			verbose = 1;
			break;
		case 'u':
		case 'q':
		case 'r':
			/* Ignored. */
			break;
		case 'h':
			print_usage(argv[0]);
			break;
		case 'n':
			depout = pciout = usbout = ccwout = stdout;
			break;
		case 'V':
			printf("%s %s\n", PACKAGE, VERSION);
			exit(0);
		default:
			bad_option = optopt;
		}
	}

	/* We can't print unknowns without a System.map */
	if (!system_map)
		print_unknown = 0;
	else
		load_system_map(system_map);

	/* They can specify the version naked on the command line */
	if (optind < argc && is_version_number(argv[optind]))
		version = strdup(argv[optind]);
	else {
		struct utsname buf;
		uname(&buf);
		version = strdup(buf.release);
	}

	/* Run old version if required. */
	if (old_module_version(version))
		exec_old_depmod(argv);

	if (bad_option) {
		fprintf(stderr, "%s: unrecognized option '%s'\n",
			argv[0], argv[bad_option]);
		print_usage(argv[0]);
		exit(1);
	}

	/* Depmod -a by default if no names. */
	if (optind == argc)
		all = 1;

	dirname = malloc(strlen(basedir)
			 + strlen("/lib/modules/")
			 + strlen(version) + 1);
	sprintf(dirname, "%s%s%s", basedir, "/lib/modules/", version);

	/* If we're not stdout, open output files. */
	if (!depout) {
		/* Longest name to allocate */
		char depname[strlen(dirname) + 1
			    + sizeof("modules.generic_string")];

		sprintf(depname, "%s/%s", dirname, "modules.dep");
		depout = fopen(depname, "w");
		if (!depout)
			fatal("Could not open %s for writing: %s\n",
			      depname, strerror(errno));

		sprintf(depname, "%s/%s", dirname, "modules.pcimap");
		pciout = fopen(depname, "w");
		if (!pciout)
			fatal("Could not open %s for writing: %s\n",
			      depname, strerror(errno));

		sprintf(depname, "%s/%s", dirname, "modules.usbmap");
		usbout = fopen(depname, "w");
		if (!usbout)
			fatal("Could not open %s for writing: %s\n",
			      depname, strerror(errno));

		sprintf(depname, "%s/%s", dirname, "modules.ccwmap");
		ccwout = fopen(depname, "w");
		if (!ccwout)
			fatal("Could not open %s for writing: %s\n",
			      depname, strerror(errno));
	}

	if (!all) {
		/* Do command line args. */
		for (opt = optind + 1; opt < argc; opt++)
			list = grab_module("", argv[opt], list);
	} else {
		list = grab_dir(dirname, list);
	}

	output_deps(list, skipchars, depout, verbose);
	output_pci_table(list, pciout);
	output_usb_table(list, usbout);
	output_ccw_table(list, ccwout);

	free(dirname);
	free(version);
	
	return 0;
}
